"""CouchDB Backend Module for django-ajaxuploader

"""

from uuid import uuid4
from tempfile import TemporaryFile

from couchdb import Server
from django.conf import settings

from ajaxuploader.backends.base import AbstractUploadBackend

class CouchDBUploadBackend(AbstractUploadBackend):
    """Stores the file in a CouchDB Backend

    Requirements: couchdb

    Settings:
    AJAXUPLOAD_COUCHDB_HOST:  Specify the host of your MongoDB server.
     Defaults to http://localhost:5984 if not specified.

    Arguments:
    db (required): Specify the database within CouchDB you wish to use

    Context returned:
    None
    
    """
    def __init__(self, *args, **kwargs):
        self.database = kwargs.pop('db')
        self.connection = None
        self._dest = None
        
        super(CouchDBUploadBackend, self).__init__(*args, **kwargs)

    def setup(self, filename, *args, **kwargs):
        self.connection = Server(getattr(settings,
                                         'AJAXUPLOAD_COUCHDB_HOST',
                                         'http://localhost:5984')
        )[self.database]
        self._dest = TemporaryFile()

    def upload_chunk(self, chunk, *args, **kwargs):
        self._dest.write(chunk)

    def upload_complete(self, request, filename, *args, **kwargs):
        self._dest.seek(0)
        doc_id = uuid4().hex
        # create doc by self defined uuid. We need the _rev for attachment
        doc = self.connection[doc_id] = {'_id':doc_id,
                                         # append anything you like maybe from
                                         # request
                                         }
        # mimetype is guessed by extension.
        # We don't put the whole document back in the put_attachment request
        self.connection.put_attachment({'_id':doc['_id'], '_rev':doc['_rev']},
                                        self._dest,
                                        filename=filename)
        self._dest.close()



import os

from django.conf import settings
from easy_thumbnails.files import get_thumbnailer

from ajaxuploader.backends.local import LocalUploadBackend

class EasyThumbnailUploadBackend(LocalUploadBackend):
    DIMENSIONS = (100,000)
    CROP = True
    KEEP_ORIGINAL = False

    def upload_complete(self, request, filename):

        options = {'size': self.DIMENSIONS, 'crop': self.CROP}
        thumb = get_thumbnailer(self._path).get_thumbnail(options)

        if not self.KEEP_ORIGINAL:
            os.unlink(self._path)

        return {"path": settings.MEDIA_URL + os.path.split(thumb.path)[1]}


# Via Fine-uploader's server examples.
# MIT License https://github.com/Widen/fine-uploader-server/blob/master/license.txt
from django.conf import settings
from django.http import HttpResponse
from django.shortcuts import render
from django.views.decorators.csrf import csrf_exempt

import base64, hmac, hashlib, json

try:
    import boto
    from boto.s3.connection import Key, S3Connection
    boto.set_stream_logger('boto')
    S3 = S3Connection(settings.AWS_UPLOAD_CLIENT_KEY, settings.AWS_UPLOAD_CLIENT_SECRET_KEY)
except ImportError, e:
    print("Could not import boto, the Amazon SDK for Python.")
    print("Deleting files will not work.")
    print("Install boto with")
    print("$ pip install boto")


def home(request):
    """ The 'home' page. Returns an HTML page with Fine Uploader code
    ready to upload to S3.
    """
    return render(request, "index.html")


@csrf_exempt
def success_redirect_endpoint(request):
    """ This is where the upload will snd a POST request after the
    file has been stored in S3.
    """
    return make_response(200)


@csrf_exempt
def handle_s3(request):
    """ View which handles all POST and DELETE requests sent by Fine Uploader
    S3. You will need to adjust these paths/conditions based on your setup.
    """
    if request.method == "POST":
        return handle_POST(request)
    elif request.method == "DELETE":
        return handle_DELETE(request)
    else:
        return HttpResponse(status=405)


def handle_POST(request):
    """ Handle S3 uploader POST requests here. For files <=5MiB this is a simple
    request to sign the policy document. For files >5MiB this is a request
    to sign the headers to start a multipart encoded request.
    """
    if request.POST.get('success', None):
        return make_response(200)
    else:
        request_payload = json.loads(request.body)
        headers = request_payload.get('headers', None)
        if headers:
            # The presence of the 'headers' property in the request payload
            # means this is a request to sign a REST/multipart request
            # and NOT a policy document
            response_data = sign_headers(headers)
        else:
            response_data = sign_policy_document(request_payload)
        response_payload = json.dumps(response_data)
        return make_response(200, response_payload)


def handle_DELETE(request):
    """ Handle file deletion requests. For this, we use the Amazon Python SDK,
    boto.
    """
    if boto:
        bucket_name = request.REQUEST.get('bucket')
        key_name = request.REQUEST.get('key')
        aws_bucket = S3.get_bucket(bucket_name, validate=False)
        aws_key = Key(aws_bucket, key_name)
        aws_key.delete()
        return make_response(200)
    else:
        return make_response(500)


def make_response(status=200, content=None):
    """ Construct an HTTP response. Fine Uploader expects 'application/json'.
    """
    response = HttpResponse()
    response.status_code = status
    response['Content-Type'] = "application/json"
    response.content = content
    return response


def is_valid_policy(policy_document):
    """ Verify the policy document has not been tampered with client-side
    before sending it off.
    """
    #bucket = settings.AWS_UPLOAD_BUCKET_NAME
    #parsed_max_size = settings.AWS_MAX_SIZE
    bucket = ''
    parsed_max_size = 0

    for condition in policy_document['conditions']:
        if isinstance(condition, list) and condition[0] == 'content-length-range':
            parsed_max_size = condition[2]
        else:
            if condition.get('bucket', None):
                bucket = condition['bucket']

    return bucket == settings.AWS_UPLOAD_BUCKET_NAME and parsed_max_size == settings.AWS_MAX_SIZE


def sign_policy_document(policy_document):
    """ Sign and return the policy doucument for a simple upload.
    http://aws.amazon.com/articles/1434/#signyours3postform
    """
    policy = base64.b64encode(json.dumps(policy_document))
    signature = base64.b64encode(hmac.new(settings.AWS_UPLOAD_CLIENT_SECRET_KEY, policy, hashlib.sha1).digest())
    return {
        'policy': policy,
        'signature': signature
    }


def sign_headers(headers):
    """ Sign and return the headers for a chunked upload. """
    return {
        'signature': base64.b64encode(hmac.new(settings.AWS_UPLOAD_CLIENT_SECRET_KEY, headers, hashlib.sha1).digest())
    }



import os
import hashlib
from shutil import rmtree

from django.conf import settings
from django.test import TestCase
from django.test.client import Client
from django.core.urlresolvers import reverse


class AjaxUploadTest(TestCase):
    urls = 'ajaxuploader.tests.urls'

    def setUp(self):
        super(AjaxUploadTest, self).setUp()
        
        self.test_dir = os.path.dirname(__file__)
        test_file = open(os.path.join(self.test_dir, '../fixtures/pony.png'),
                         'rb')

        self.test_file_1 = test_file

        # generate sha1 hash of the file
        self.test_file_1_hash = hashlib.sha1(self.test_file_1.read())\
                                       .hexdigest()

        # reset position to beginning of the file
        self.test_file_1.seek(0)
        
        
    def tearDown(self):
        # remove created uploads/tests directory
        rmtree(os.path.join(settings.MEDIA_ROOT, 'uploads/tests'))
        
    def test_upload_raw_post_local_backend(self):
        """
        tests uploading a file to DefaultStorageUploadBackend
        """

        # there is a bug in Django 1.3.1 with raw_post_data in the django test client
        # it's fixed in trunk
        # https://code.djangoproject.com/changeset/16479

        uploaded_file_name = 'tests/foo.png'

        file_data = self.test_file_1.read()
        # post raw self.test_file_1 data to AjaxFileUploader as Ajax request
        response = self.client.post(reverse('ajax-upload-default-storage')+'?qqfile=%s' % \
                                    uploaded_file_name, file_data,
                                    content_type='application/octet-stream',
                                    HTTP_X_REQUESTED_WITH='XMLHttpRequest')

        uploaded_file_path = os.path.join(settings.MEDIA_ROOT, 'uploads',
                                          uploaded_file_name)
        uploaded_file = open(uploaded_file_path, 'rb')

        # uploaded file must exist in MEDIA_DIR
        self.assertTrue(os.path.exists(uploaded_file_path))

        # sha1 hash of original file and uploaded file must match
        self.assertEquals(hashlib.sha1(uploaded_file.read()).hexdigest(),
                          self.test_file_1_hash)



from django.conf.urls.defaults import *

from ajaxuploader.views import AjaxFileUploader
from ajaxuploader.backends.default_storage import DefaultStorageUploadBackend


default_storage_uploader = AjaxFileUploader(backend=DefaultStorageUploadBackend)

urlpatterns = patterns('',
    url(r'^upload$', default_storage_uploader, name="ajax-upload-default-storage"),                
)



import os

from django.core.files.storage import default_storage
from django.core.files.base import ContentFile

from ajaxuploader.backends.base import AbstractUploadBackend


class DefaultStorageUploadBackend(AbstractUploadBackend):
    """
    Uses Django's default storage backend to store the uploaded files
    see https://docs.djangoproject.com/en/dev/topics/files/#file-storage
    """
    
    UPLOAD_DIR = 'uploads'

    def setup(self, filename, *args, **kwargs):
        # join UPLOAD_DIR with filename 
        new_path = os.path.join(self.UPLOAD_DIR, filename)

        # save empty file in default storage with path = new_path
        self.path = default_storage.save(new_path, ContentFile(''))

        # create BufferedWriter for new file
        self._dest = default_storage.open(self.path, mode='wb')

    def upload_chunk(self, chunk, *args, **kwargs):
        self._dest.write(chunk)

    def upload_complete(self, request, filename, *args, **kwargs):
        self._dest.close()
        return {"path": self.path}



try:
    import json
except ImportError:
    from django.utils import simplejson as json
from django.core.serializers.json import DjangoJSONEncoder

from django.http import HttpResponse, HttpResponseBadRequest, Http404, HttpResponseNotAllowed

from ajaxuploader.backends.local import LocalUploadBackend

class AjaxFileUploader(object):
    def __init__(self, backend=None, **kwargs):
        if backend is None:
            backend = LocalUploadBackend
        self.get_backend = lambda: backend(**kwargs)

    def __call__(self, request, *args, **kwargs):
        return self._ajax_upload(request, *args, **kwargs)

    def _ajax_upload(self, request, *args, **kwargs):
        if request.method == "POST":
            if request.is_ajax():
                # the file is stored raw in the request
                upload = request
                is_raw = True
                # AJAX Upload will pass the filename in the querystring if it
                # is the "advanced" ajax upload
                try:
                    filename = request.GET['qqfile']
                except KeyError:
                    return HttpResponseBadRequest("AJAX request not valid")
            # not an ajax upload, so it was the "basic" iframe version with
            # submission via form
            else:
                is_raw = False
                if len(request.FILES) == 1:
                    # FILES is a dictionary in Django but Ajax Upload gives
                    # the uploaded file an ID based on a random number, so it
                    # cannot be guessed here in the code. Rather than editing
                    # Ajax Upload to pass the ID in the querystring, observe
                    # that each upload is a separate request, so FILES should
                    # only have one entry. Thus, we can just grab the first
                    # (and only) value in the dict.
                    upload = request.FILES.values()[0]
                else:
                    raise Http404("Bad Upload")
                filename = upload.name

            backend = self.get_backend()

            # custom filename handler
            filename = (backend.update_filename(request, filename, *args, **kwargs)
                        or filename)
            # save the file
            backend.setup(filename, *args, **kwargs)
            success = backend.upload(upload, filename, is_raw, *args, **kwargs)
            # callback
            extra_context = backend.upload_complete(request, filename, *args, **kwargs)

            # let Ajax Upload know whether we saved it or not
            ret_json = {'success': success, 'filename': filename}
            if extra_context is not None:
                ret_json.update(extra_context)

            # although "application/json" is the correct content type, IE throws a fit
            return HttpResponse(json.dumps(ret_json, cls=DjangoJSONEncoder), content_type='text/html; charset=utf-8')
        else:
            response = HttpResponseNotAllowed(['POST'])
            response.write("ERROR: Only POST allowed")
            return response



from django.conf.urls import patterns, url

from .views import handle_s3, success_redirect_endpoint

urlpatterns = patterns(
    '',
    url(r'^s3/signature', handle_s3, name="s3_signature"),
    url(r'^s3/delete', handle_s3, name='s3_delete'),
    url(r'^s3/success', success_redirect_endpoint, name="s3_success")
)



import os

from django.conf import settings
from sorl.thumbnail import get_thumbnail

from ajaxuploader.backends.local import LocalUploadBackend

class ThumbnailUploadBackend(LocalUploadBackend):
    DIMENSIONS = "100x100"
    KEEP_ORIGINAL = False

    def upload_complete(self, request, filename, *args, **kwargs):
        thumbnail = get_thumbnail(self._path, self.DIMENSIONS)
        if not self.KEEP_ORIGINAL:
            os.unlink(self._path)
        return {"path": settings.MEDIA_URL + thumbnail.name}



from multiprocessing import Pool
from StringIO import StringIO

import boto
from django.conf import settings

from ajaxuploader.backends.base import AbstractUploadBackend


class S3UploadBackend(AbstractUploadBackend):
    NUM_PARALLEL_PROCESSES = 4

    def upload_chunk(self, chunk, *args, **kwargs):
        self._counter += 1
        buffer = StringIO()
        buffer.write(chunk)
        self._pool.apply_async(
            self._mp.upload_part_from_file(buffer, self._counter))
        buffer.close()

    def setup(self, filename, *args, **kwargs):
        self._bucket = boto.connect_s3(
            settings.AWS_ACCESS_KEY_ID,
            settings.AWS_SECRET_ACCESS_KEY
        ).lookup(settings.AWS_BUCKET_NAME)

        self._mp = self._bucket.initiate_multipart_upload(filename)
        self._pool = Pool(processes=self.NUM_PARALLEL_PROCESSES)
        self._counter = 0

    def upload_complete(self, request, filename, *args, **kwargs):
        # Tie up loose ends, and finish the upload
        self._pool.close()
        self._pool.join()
        self._mp.complete_upload()



#/usr/bin/env python
import os
from setuptools import setup, find_packages

ROOT_DIR = os.path.dirname(__file__)
SOURCE_DIR = os.path.join(ROOT_DIR)

setup(
    name="ajaxuploader",
    description="AJAX file uploader for django",
    author="Steven Skoczen",
    author_email="steven@agoodcloud.com",
    url="https://github.com/GoodCloud/django-ajax-uploader",
    version="0.3.1",
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,  # because we're including media that Django needs
)



class AbstractUploadBackend(object):
    BUFFER_SIZE = 10485760  # 10MB

    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)

    def setup(self, filename, *args, **kwargs):
        """Responsible for doing any pre-processing needed before the upload
        starts."""

    def update_filename(self, request, filename, *args, **kwargs):
        """Returns a new name for the file being uploaded."""

    def upload_chunk(self, chunk, *args, **kwargs):
        """Called when a string was read from the client, responsible for 
        writing that string to the destination file."""
        raise NotImplementedError

    def upload_complete(self, request, filename, *args, **kwargs):
        """Overriden to performs any actions needed post-upload, and returns
        a dict to be added to the render / json context"""

    def upload(self, uploaded, filename, raw_data, *args, **kwargs):
        try:
            if raw_data:
                # File was uploaded via ajax, and is streaming in.
                chunk = uploaded.read(self.BUFFER_SIZE)
                while len(chunk) > 0:
                    self.upload_chunk(chunk, *args, **kwargs)
                    chunk = uploaded.read(self.BUFFER_SIZE)
            else:
                # File was uploaded via a POST, and is here.
                for chunk in uploaded.chunks():
                    self.upload_chunk(chunk, *args, **kwargs)
            return True
        except:
            # things went badly.
            return False



import mimetypes

from django.conf import settings

from pymongo import Connection
import gridfs

from ajaxuploader.backends.base import AbstractUploadBackend


class MongoDBUploadBackend(AbstractUploadBackend):
    """
    Stores uploaded file in Mongo's GridFS.

    Requirements: pymongo
    """
    def __init__(self, *args, **kwargs):
        self.db = kwargs.pop('db')

        try:
            self.collection = kwargs.pop('collection')
        except KeyError:
            self.collection = None

        super(MongoDBUploadBackend, self).__init__(*args, **kwargs)

    def setup(self, filename, encoding='UTF-8', *args, **kwargs):
        """
        Setup MongoDB connection to specified db. Collection is optional
        and will default to fs if not specified.

        Create new gridFS file which returns a GridIn instance to write
        data to.
        """
        host = getattr(settings, "AJAXUPLOAD_MONGODB_HOST", "localhost:27017")
        port = getattr(settings, "AJAXUPLOAD_MONGODB_PORT", 27017)
        replicaset = getattr(settings, "AJAXUPLOAD_MONGODB_REPLICASET", "")

        if isinstance(host, list):
            host = ", ".join(host)
        else:
            if not ":" in host:
                """
                Backwards compatibility for old version.
                """
                host = u"%s:%d" % (host, port)
        self.connection = Connection(host, replicaset=replicaset)[self.db]

        if self.collection:
            self.grid = gridfs.GridFS(self.connection,
                collection=self.collection)
        else:
            self.grid = gridfs.GridFS(self.connection)

        self._path = filename
        content_type, encoding = mimetypes.guess_type(filename)
        self.f = self.grid.new_file(**{'filename': self._path,
            'encoding': 'UTF-8', 'content_type': content_type})

    def upload_chunk(self, chunk, *args, **kwargs):
        self.f.write(chunk)

    def upload_complete(self, request, filename, *args, **kwargs):
        self.f.close()
